import tomli
from .core import ApiProxy

# Python Standard Library
import filecmp, io, os, subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable, Optional
from warnings import warn


class Config:
    @dataclass
    class Task:
        settings: Any
        react: list[str]
        clean: bool

    def __init__(self, config_file: Path):
        with open(config_file, "rb") as file:
            self._data = tomli.load(file)
        self.path = config_file

    def _resolve_path(self, s: Any) -> Optional[Path]:
        ret = None
        if s is not None:
            ret = Path(str(s)).expanduser()
            if not ret.is_absolute():
                ret = self.path.parent / ret
        return ret

    def get_api_proxy(self, log_path: Path) -> ApiProxy:
        api_key = None
        api_key_path = self._data.get("openai_api_key_file")
        api_key_path = self._resolve_path(api_key_path)
        if api_key_path is not None:
            with open(api_key_path, 'r') as file:
                api_key = file.read().strip()
        return ApiProxy(api_key, log_path, self.log_format)

    @property
    def log_format(self) -> Optional[str]:
        return self._data.get("log_format")

    @property
    def task_names(self) -> Iterable[str]:
        tasks = self._data.get("tasks", {})
        assert isinstance(tasks, dict)
        return tasks.keys()

    def _react_as_commands(self, react: Any) -> list[str]:
        ret = list()
        if react is None:
            react = []
        elif isinstance(react, str):
            react = [react]
        for r in react:
            cmd = self._data.get("commands", {}).get(r)
            if cmd is None:
                msg = f"Command '{r}' not found in {self.path}"
                raise SyntaxError(msg)
            ret.append(cmd)
        return ret

    def get_task(self, task_name: str) -> Optional[Task]:
        task = self._data.get("tasks", {}).get(task_name)
        if task is None:
            return None
        path = self._resolve_path(task.get("request"))
        if path is None:
            settings = None
        else:
            with open(path, "rb") as file:
                settings = tomli.load(file)
        react = self._react_as_commands(task.get("react"))
        clean = task.get("clean", False)
        return Config.Task(settings, react, clean)

    def help(self) -> str:
        buf = io.StringIO()
        buf.write("task choices:\n")
        for name in self.task_names:
            buf.write("  ")
            buf.write(name)
            buf.write("\n")
            task = self._data["tasks"][name]
            path = self._resolve_path(task.get("request"))
            if path:
                if task.get("clean"):
                    buf.write("    Clean & request: ")
                else:
                    buf.write("    Request: ")
                buf.write(str(path))
                buf.write("\n")
            react = self._react_as_commands(task.get("react"))
            if react:
                buf.write("    React command chain:\n")
                for r in react:
                    buf.write("      ")
                    buf.write(help_example_react(r))
            buf.write("\n")
        return buf.getvalue()


def help_example_react(cmd: str) -> str:
    subs = {
        '"$0"': "<source>",
        '"$1"': "<rev1>",
        '"$@"': "<rev1> ... <revN>",
    }
    for k, v in subs.items():
        cmd = cmd.replace(k, v)
    return cmd + "\n"

class Task:
    MAX_NUM_REVS = 7

    def __init__(self, dest: Path, config: Config, task_name: str):
        tconfig = config.get_task(task_name)
        if tconfig is None:
            raise ValueError(f"Invalid task name {task_name}.")
        self.dest = dest
        self.settings = tconfig.settings
        self.react = tconfig.react
        self.clean = tconfig.clean
        if self.settings is not None:
            num_revs = self.settings.get("n", 1)
            if num_revs > Task.MAX_NUM_REVS:
                self.settings["n"] = Task.MAX_NUM_REVS
                msg = "{} revisions requested, changed to max {}"
                warn(msg.format(num_revs, Task.MAX_NUM_REVS))
        if self.clean and not self.settings:
            self.clean = False
            msg = "Setting clean=true ignored for task {} since there is no request."
            warn(msg.format(task_name))

    def _rev_paths(self, src_path: Path) -> list[Path]:
        ret = list()
        pattern = str(self.dest) + "/R{}/" + src_path.name
        for i in range(Task.MAX_NUM_REVS):
            ret.append(Path(pattern.format(i + 1)))
        return ret

    def use_saved_revision(self, src: Path) -> Optional[Path]:
        if not self.clean:
            R1 = self.dest / "R1" / src.name
            R2 = self.dest / "R2" / src.name
            if R1.exists() and not R2.exists():
                if filecmp.cmp(src, R1, shallow=False):
                    return R1
        return None

    def write_revisions(self, src_path: Path, revisions: list[str]) -> None:
        for i, path in enumerate(self._rev_paths(src_path)):
            if i < len(revisions):
                os.makedirs(path.parent, exist_ok=True)
                with open(path, "w") as file:
                    file.write(revisions[i])
            else:
                path.unlink(missing_ok=True)

    def do_react(self, src_path: Path) -> int:
        ret = 0
        if self.react is not None:
            found_revs = [str(p) for p in self._rev_paths(src_path) if p.exists()]
            if found_revs:
                args = [str(src_path)] + found_revs
                for cmd in self.react:
                    proc = subprocess.run([cmd] + args, shell=True)
                    ret = proc.returncode
                    if ret:
                        break
        return ret



# Python standard libraries
import os, shutil
from importlib import resources
from pathlib import Path

STD_BASE_DIRS = dict(
    TMPDIR="/tmp",
    XDG_CONFIG_HOME="~/.config",
    XDG_STATE_HOME="~/.local/state",
)


def get_std_path(env_var_name: str, subpath: str) -> Path:
    base_dir = os.environ.get(env_var_name)
    if base_dir is None:
        base_dir = STD_BASE_DIRS[env_var_name]
    return Path(base_dir).expanduser() / subpath


def copy_package_file(filename: str, dest: Path) -> None:
    rp = resources.files(__package__).joinpath("data").joinpath(filename)
    with resources.as_file(rp) as filepath:
        os.makedirs(dest.parent, exist_ok=True)
        shutil.copy(filepath, dest)



from .core import ApiProxy
from .task import Config, Task
from .util import get_std_path, copy_package_file

# Python standard libraries
import argparse
from pathlib import Path
from sys import stderr
from typing import Optional, Union

COPYAID_TMP_DIR = ("TMPDIR", "copyaid")
COPYAID_CONFIG_FILENAME = "copyaid.toml"
COPYAID_CONFIG_FILE = ("XDG_CONFIG_HOME", "copyaid/" + COPYAID_CONFIG_FILENAME)
COPYAID_LOG_DIR = ("XDG_STATE_HOME", "copyaid/log")


def preparse_config(prog: str, cmd_line_args: Optional[list[str]]) -> Union[int, Path]:
    preparser = argparse.ArgumentParser(add_help=False)
    preparser.add_argument("-c", "--config", type=Path)
    (args, rest) = preparser.parse_known_args(cmd_line_args)
    config_path = args.config or Path(get_std_path(*COPYAID_CONFIG_FILE))
    if config_path.is_dir():
        config_path = config_path / COPYAID_CONFIG_FILENAME
    if not config_path.exists():
        if rest == ["init"]:
            copy_package_file(COPYAID_CONFIG_FILENAME, config_path)
            copy_package_file("cold-example.toml", config_path.parent)
            return 0
        else:
            print(f"Config file '{config_path}' not found, run:", file=stderr)
            if args.config is None:
                print(f"  {prog} init", file=stderr)
            else:
                print(f"  {prog} --config '{args.config}' init", file=stderr)
            return 2
    return config_path


def main(cmd_line_args: Optional[list[str]] = None) -> int:
    prog = "copyaid"
    ret = preparse_config(prog, cmd_line_args)
    if isinstance(ret, int):
        return ret
    config = Config(ret)
    parser = argparse.ArgumentParser(
        prog=prog,
        description="CopyAId",
        epilog=config.help(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--config",
        default=argparse.SUPPRESS,
        metavar="<config>",
        help="Configuration file"
    )
    parser.add_argument(
        "-d",
        "--dest",
        type=Path,
        metavar="<dest>",
        help="Destination directory for revisions"
    )
    parser.add_argument("task", choices=config.task_names, metavar="<task>")
    parser.add_argument("source", type=Path, nargs="+", metavar="<source>")
    args = parser.parse_args(cmd_line_args)
    if args.dest is None:
        args.dest = Path(get_std_path(*COPYAID_TMP_DIR))
    ret = check_filename_collision(args.source)
    if ret == 0:
        ret = do_task(config, args.task, args.source, args.dest)
    return ret


def check_filename_collision(sources: list[Path]) -> int:
    filenames = set()
    for s in sources:
        if s.name in filenames:
            msg = "Sources must have unique filenames. Conflict: {}"
            print(msg.format(s.name), file=stderr)
            return 2
        filenames.add(s.name)
    return 0


def do_task(config: Config, task_name: str, sources: list[Path], dest: Path) -> int:
    exit_code = 0
    task = Task(dest, config, task_name)
    if task.settings is None:
        api = None
    else:
        api = config.get_api_proxy(get_std_path(*COPYAID_LOG_DIR))
    for s in sources:
        if not s.exists():
            print(f"File not found: '{s}'", file=stderr)
            exit_code = 2
            break
        if api:
            saved = task.use_saved_revision(s)
            if saved:
                print("Reusing saved", saved)
            else:
                print("OpenAI request for", s)
                revisions = api.do_request(task.settings, s)
                print("Saving to", task.dest)
                task.write_revisions(s, revisions)
        exit_code |= task.do_react(s)
        if exit_code > 1:
            break
    return exit_code



# Python Standard Library
import json, os
from datetime import datetime
from pathlib import Path
from typing import Any, Optional
from warnings import warn


class LiveOpenAiApi:
    def __init__(self, api_key: Optional[str] = None):
        import openai  # delay a very slow import

        if api_key is not None:
            openai.api_key = api_key

    def query(self, req: Any) -> Any:
        import openai

        return openai.ChatCompletion.create(**req)  # type: ignore


def make_openai_request(settings: Any, source: str) -> Any:
    ret = settings["openai"]
    ret["max_tokens"] = max(6, int(settings["max_tokens_ratio"] * len(source) / 4))
    prompt = settings.get("prepend", "") + source + settings.get("append", "")
    ret["messages"] = [
        {
            "role": "system",
            "content": settings["chat_system"]
        },
        {
            "role": "user",
            "content": prompt
        },
    ]
    return ret


class ApiProxy:
    ApiClass = LiveOpenAiApi

    def __init__(
        self, api_key: Optional[str], log_path: Path, log_format: Optional[str]
    ):
        self.log_path = log_path
        self.log_format = log_format
        self._api = ApiProxy.ApiClass(api_key)

    def do_request(self, settings: Any, src_path: Path) -> list[str]:
        with open(src_path) as file:
            source_text = file.read()
        request = make_openai_request(settings, source_text)
        response = self._api.query(request)
        self.log_openai_query(src_path.stem, request, response)
        revisions = list()
        for choice in response.get("choices", []):
            text = choice.get("message", {}).get("content")
            revisions.append(text)
        return revisions

    def log_openai_query(self, name: str, request: Any, response: Any) -> None:
        if not self.log_format:
            return
        t = datetime.utcfromtimestamp(response["created"])
        ts = t.isoformat().replace("-", "").replace(":", "") + "Z"
        data = dict(request=request, response=response)
        os.makedirs(self.log_path, exist_ok=True)
        save_stem = name + "." + ts
        print("Logging OpenAI response", save_stem)
        if self.log_format == "jsoml":
            import jsoml
            jsoml.dump(data, self.log_path / (save_stem + ".xml"))
        elif self.log_format == "json":
            with open(self.log_path / (save_stem + ".json"), "w") as file:
                json.dump(data, file, indent=4, ensure_ascii=False)
                file.write("\n")
        else:
            warn("Unsupported log format: {}".format(self.log_format))



import pytest

import copyaid.cli


SOURCE_TEXT = "Jupiter big.\nJupiter a planet.\nJupiter gas.\n"
MOCK_COMPLETION = "Jupiter is a big planet made of gas."
EXPECTED_TEXT = "Jupiter is\na big planet made of\ngas.\n"


class MockApi:
    def __init__(self, api_key):
        pass

    def query(self, req):
        ret = dict(
            created=1674259148,
            choices=[dict(message=dict(content=MOCK_COMPLETION))],
        )
        return ret

copyaid.core.ApiProxy.ApiClass = MockApi

def test_main(tmp_path):
    srcpath = tmp_path / "source.txt"
    open(srcpath, "w").write(SOURCE_TEXT)
    retcode = copyaid.cli.main([
        "proof",
        str(srcpath),
        "--dest", str(tmp_path),
        "--config", "tests/mock_config.toml",
    ])
    assert retcode == 0
    got = open(tmp_path / "R1" / srcpath.name).read()
    assert got == EXPECTED_TEXT
